#!/usr/bin/python

################################################################
#                                                              #
# Blue Team Training Toolkit (BT3)                             #
# written by Juan J. Guelfo @ Encripto AS                      #
# post@encripto.no                                             #
#                                                              #
# Copyright 2013-2016 Encripto AS. All rights reserved.        #
#                                                              #
# BT3 is licensed under the FreeBSD license.                   #
# http://www.freebsd.org/copyright/freebsd-license.html        #
#                                                              #
################################################################


import cmd, os, sys, re, ipcalc, subprocess, datetime, signal
import libs.bt3out, libs.bt3in

import logging
logging.getLogger("scapy.runtime").setLevel(logging.ERROR)
from scapy.all import *


__module_name__    = "Pcapteller"
__module_version__ = "1.2"


class Options():
    
    def __init__(self):        
        self.file          = ""
        self.fragmentation = False
        self.interface     = ""
        self.mtu           = "1500"
        self.pcap_ip_list  = ""
        self.pcap_mac_list = ""
        self.real_time     = False
        self.wire_ip_list  = ""
        self.wire_mac_list = ""
        self.options       = ""
        
    def get_options(self):
        self.options = [["FILE", str(self.file), "True", "Pcap file to replay in libpcap format."],
                        ["FRAGMENTATION", str(self.fragmentation), "True", "Fragment packets during replay. Useful for networks with low MTU."],
                        ["INTERFACE", str(self.interface), "True", "Network interface to replay the packets with."],
                        ["MTU", str(self.mtu), "True", "MTU to use with packet fragmentation."],
                        ["PCAP_IP_LIST", str(self.pcap_ip_list), "False", "Comma-separated list of IP addresses to replace as seen on the pcap file."],
                        ["PCAP_MAC_LIST", str(self.pcap_mac_list), "False", "Comma-separated list of MAC addresses to replace as seen on the pcap file."],
                        ["REAL_TIME", str(self.real_time), "False", "Honor inter-packet arrival time while replaying traffic."],
                        ["WIRE_IP_LIST", str(self.wire_ip_list), "False", "Comma-separated list of IP addresses to replay as seen on the wire."],
                        ["WIRE_MAC_LIST", str(self.wire_mac_list), "False", "Comma-separated list of MAC addresses to replay as seen on the wire."]]

        return self.options


    def validate_options(self):
        print("")
        return self.validate_mandatory_args(self.file, self.fragmentation, self.mtu, self.interface) and self.validate_optional_args(self.pcap_ip_list.split(","), self.wire_ip_list.split(","), 
                                                                                                       self.pcap_mac_list.split(","), self.wire_mac_list.split(","), self.real_time)
        
        
    def validate_mandatory_args(self, pcap_file, fragmentation, mtu_bytes, net_interface):
        valid_options = True
                       
        base_path = "%s/%s" % (os.path.abspath(os.path.curdir), "pcaps")
        pcap_path = "%s/%s" % (base_path, pcap_file)
        is_valid_path = os.path.abspath(pcap_path).startswith(base_path)
        if not pcap_file or not is_valid_path or not libs.bt3in.check_file_exists(pcap_path):
            libs.bt3out.print_error("A valid pcap file (located in 'pcaps') must be provided.\n")
            valid_options = False
        
        if type(fragmentation) != type(True):
            libs.bt3out.print_error("Fragmentation option value must be boolean (True or False).\n")
            valid_options = False
        
        try:
            mtu_bytes = int(mtu_bytes)
            if mtu_bytes < 1 or mtu_bytes > 9000:
                libs.bt3out.print_error("A valid MTU size must be provided.\n")
                valid_options = False 
       
        except:
            libs.bt3out.print_error("A valid MTU size must be provided.\n")
            valid_options = False        
       
        if not net_interface or not libs.bt3in.check_network_interface(net_interface):        
            libs.bt3out.print_error("A valid network interface must be provided.\n")
            valid_options = False
         
        return valid_options


    def validate_optional_args(self, pcap_ip_addr_list, wire_ip_addr_list, pcap_mac_addr_list, wire_mac_addr_list, real_time):
        valid_options = True
        if (pcap_ip_addr_list and not wire_ip_addr_list) or (wire_ip_addr_list and not pcap_ip_addr_list):
            libs.bt3out.print_error("Lists of matching pcap and wire IP addresses must be provided.\n")
            valid_options = False
            
        if (pcap_mac_addr_list and not wire_mac_addr_list) or (wire_mac_addr_list and not pcap_mac_addr_list):
            libs.bt3out.print_error("Lists of matching pcap and wire MAC addresses must be provided.\n")
            valid_options = False
            
        if pcap_ip_addr_list and wire_ip_addr_list:
            if len(pcap_ip_addr_list) != len(wire_ip_addr_list):
                libs.bt3out.print_error("Pcap and wire IP address lists must have the same number of elements.\n")
                valid_options = False
                
            elif len(pcap_ip_addr_list) == 1 and len(wire_ip_addr_list) == 1:
                if (pcap_ip_addr_list[0] and not wire_ip_addr_list[0]) or (not pcap_ip_addr_list[0] and wire_ip_addr_list[0]):
                    libs.bt3out.print_error("Pcap and wire IP address lists must have the same number of elements.\n")
                    valid_options = False
            
            for pcap_ip_addr in pcap_ip_addr_list:
                if pcap_ip_addr and not libs.bt3in.check_ip_address(pcap_ip_addr):
                    libs.bt3out.print_error("Pcap IP address \"%s\" is not valid.\n" % (pcap_ip_addr))
                    valid_options = False
                
            for wire_ip_addr in wire_ip_addr_list:
                if wire_ip_addr and not libs.bt3in.check_ip_address(wire_ip_addr):
                    libs.bt3out.print_error("Wire IP address \"%s\" is not valid.\n" % (wire_ip_addr))
                    valid_options = False
                         
        if pcap_mac_addr_list and wire_mac_addr_list:
            if len(pcap_mac_addr_list) != len(wire_mac_addr_list):
                libs.bt3out.print_error("Pcap and wire MAC address lists must have the same number of elements.\n")
                valid_options = False

            elif len(pcap_mac_addr_list) == 1 and len(wire_mac_addr_list) == 1:
                if (pcap_mac_addr_list[0] and not wire_mac_addr_list[0]) or (not pcap_mac_addr_list[0] and wire_mac_addr_list[0]):
                    libs.bt3out.print_error("Pcap and wire MAC address lists must have the same number of elements.\n")
                    valid_options = False
                 
            for pcap_mac_addr in pcap_mac_addr_list:
                if pcap_mac_addr and not libs.bt3in.check_mac_address(pcap_mac_addr):
                    libs.bt3out.print_error("Pcap MAC address \"%s\" is not valid.\n" % (pcap_mac_addr))
                    valid_options = False   
        
            for wire_mac_addr in wire_mac_addr_list:
                if wire_mac_addr and not libs.bt3in.check_mac_address(wire_mac_addr):
                    libs.bt3out.print_error("Wire MAC address \"%s\" is not valid.\n" % (wire_mac_addr))
                    valid_options = False
                    
        if type(real_time) != type(True):
            libs.bt3out.print_error("Real time option value must be boolean (True or False).\n")
            valid_options = False
            
        return valid_options



class Menu(cmd.Cmd):

    def __init__(self):
        cmd.Cmd.__init__(self)
        self.config = Options()
        self.pcapteller = Pcapteller()


    def complete_set(self, text, line, begidx, endidx):
        return [i for i in ["file", "FILE", "fragmentation", "FRAGMENTATION", "interface", "INTERFACE", "mtu", "MTU", 
                            "pcap_ip_list", "PCAP_IP_LIST", "pcap_mac_list", "PCAP_MAC_LIST", "REAL_TIME", 
                            "real_time", "wire_ip_list", "WIRE_IP_LIST", "wire_mac_list", "WIRE_MAC_LIST" ] if i.startswith(text)]


    def complete_show(self, text, line, begidx, endidx):
        return [i for i in ["options"] if i.startswith(text)]
    
    
    def default(self, args):
        print("")
        libs.bt3out.print_error("Command not found. Type \"help\" if you need assistance.\n")


    def emptyline(self):
        pass


    def do_back(self, args):
        return True


    def do_exit(self, args):
        libs.bt3out.exit_program(args)


    def do_help(self, args):
        cmds = [["back", "Exit current selected module and return to main menu."],
                ["exit", "Exit the Blue Team Training Toolkit."],        
                ["help", "Display help menu."],
                ["run", "Run the module with the given options."],
                ["set <option> <value>", "Set module option."],
                ["show options", "Display module options."],
                ["version", "Display module version."]]
        libs.bt3out.print_help(cmds, False, True)


    def do_set(self, args):
        try:
            args = args.split(" ")
            option_set = False
            if len(args) == 2:                        
                if args[0].upper() in ["FILE"]:
                    self.config.file = os.path.basename(args[1])
                    option_set = True
                    
                elif args[0].upper() in ["FRAGMENTATION"]:
                    if args[1].upper() in ["TRUE"]:
                        self.config.fragmentation = True
                    else:
                        self.config.fragmentation = False
        
                    option_set = True
                    
                elif args[0].upper() in ["INTERFACE"]:
                    self.config.interface = args[1]
                    option_set = True
                    
                elif args[0].upper() in ["MTU"]:
                    self.config.mtu = args[1]
                    option_set = True

                elif args[0].upper() in ["PCAP_IP_LIST"]:
                    if args[1].upper() in ["\"\""]:
                        self.config.pcap_ip_list = ""
                    else:
                        self.config.pcap_ip_list = args[1]
                        
                    option_set = True
                    
                elif args[0].upper() in ["PCAP_MAC_LIST"]:
                    if args[1].upper() in ["\"\""]:
                        self.config.pcap_mac_list = ""
                    else:
                        self.config.pcap_mac_list = args[1]
                        
                    option_set = True
                    
                elif args[0].upper() in ["REAL_TIME"]:
                    if args[1].upper() in ["TRUE"]:
                        self.config.real_time = True
                    else:
                        self.config.real_time = False
                        
                    option_set = True

                elif args[0].upper() in ["WIRE_IP_LIST"]:
                    if args[1].upper() in ["\"\""]:
                        self.config.wire_ip_list = ""                
                    else:
                        self.config.wire_ip_list = args[1]
    
                    option_set = True
                    
                elif args[0].upper() in ["WIRE_MAC_LIST"]:
                    if args[1].upper() in ["\"\""]:
                        self.config.wire_mac_list = ""
                    else:
                        self.config.wire_mac_list = args[1]
                        
                    option_set = True

            if option_set:
                print("")
                if args[0] in ["FILE"]:
                    libs.bt3out.print_ok("%s => %s\n" % (args[0], os.path.basename(args[1])))                
                else:
                    libs.bt3out.print_ok("%s => %s\n" % (args[0], args[1]))
                    
            else:
                cmds = [["set FILE <file name>", "Set a pcap file to replay in libpcap format."],
                        ["set FRAGMENTATION <bool>", "Set whether packets should be fragmented during replay. Useful for networks with low MTU."],
                        ["set INTERFACE <net iface>", "Set a network interface to replay the packets with."],
                        ["set MTU <int>", "Set the packet fragmentation MTU size."],
                        ["set PCAP_IP_LIST <ip addr>", "Set a comma-separated list of IP addresses to replace as seen on the pcap file."],
                        ["set PCAP_MAC_LIST <mac addr>", "Set a comma-separated list of MAC addresses to replace as seen on the pcap file."],
                        ["set REAL_TIME <bool>", "Set whether inter-packet arrival time should be honored while replaying traffic."],
                        ["set WIRE_IP_LIST <ip addr>", "Set a comma-separated list of IP addresses to replay as seen on the wire."],
                        ["set WIRE_MAC_LIST <mac addr>", "Set a comma-separated list of MAC addresses to replay as seen on the wire."],
                        ]
                libs.bt3out.print_help(cmds, True, False)

        except Exception as e:
            print("")
            libs.bt3out.print_error("%s\n" % e)        


    def do_show(self, args):
        if args.upper() in ["OPTIONS"]:
            libs.bt3out.print_options(self.config.get_options())

        else:
            cmds = [["show options", "Display module options"]]
            libs.bt3out.print_help(cmds, True, False)


    def do_run(self, args):
        if self.config.validate_options():
            base_path = "%s/%s" % (os.path.abspath(os.path.curdir), "pcaps")
            pcap_path = "%s/%s" % (base_path, self.config.file)
            self.pcapteller.run(pcap_path, self.config.fragmentation, self.config.mtu, self.config.interface, self.config.pcap_ip_list.split(","), 
                                self.config.wire_ip_list.split(","), self.config.pcap_mac_list.split(","), self.config.wire_mac_list.split(","), self.config.real_time)
        

    def do_version(self, args):
        print("")
        libs.bt3out.print_info("You are running %s version %s\n" % (__module_name__, __module_version__))


        
class Pcapteller():

    def read_traffic(self, pcap_file):
        return rdpcap(pcap_file)
    

    def traffic_manipulation(self, pcap_file, total_packets, pcap_ip_addr_list, wire_ip_addr_list, pcap_mac_addr_list, wire_mac_addr_list):
        packet_counter = 0
        error_counter = 0
        packets = []
        for packet in pcap_file:
            try:
                packet_counter += 1
                sys.stdout.write("\r\033[1;34m[*]\033[1;m Processing %s of %s packet(s) | Error: %s packet(s)." % (packet_counter, total_packets, error_counter))
                sys.stdout.flush()
                
                if packet.haslayer(Ether):
                    del(packet[Ether].chksum)
                    if pcap_mac_addr_list and wire_mac_addr_list:
                        for pcap_mac_addr, wire_mac_addr in zip(pcap_mac_addr_list, wire_mac_addr_list):
                            if packet[Ether].src == pcap_mac_addr:
                                packet[Ether].src = wire_mac_addr

                            elif packet[Ether].dst == pcap_mac_addr:
                                packet[Ether].dst = wire_mac_addr

                if packet.haslayer(UDP):
                    del(packet[UDP].chksum)

                if packet.haslayer(ICMP):
                    del(packet[ICMP].chksum)

                if packet.haslayer(TCP):
                    del(packet[TCP].chksum)

                if packet.haslayer(IP):
                    del(packet[IP].chksum)
                    if pcap_ip_addr_list and wire_ip_addr_list:
                        for pcap_ip_addr, wire_ip_addr in zip(pcap_ip_addr_list, wire_ip_addr_list):
                            if packet[IP].src == pcap_ip_addr:
                                packet[IP].src = wire_ip_addr

                            elif packet[IP].dst == pcap_ip_addr:
                                packet[IP].dst = wire_ip_addr
                        
                packets.append(packet)

            except:
                error_counter += 1
        
        return packets


    def traffic_replay(self, packets, net_interface, real_time, fragmentation, mtu_bytes):
        error = False
        try:
            if fragmentation:
                sendp(fragment(packets, int(mtu_bytes)), iface=net_interface, verbose=0, realtime=real_time)        
            else:
                sendp(packets, iface=net_interface, verbose=0, realtime=real_time)

        except:
            error = True
            
        return error
        
    
    def run(self, pcap_file, fragmentation, mtu_bytes, net_interface, pcap_ip_addr_list, wire_ip_addr_list, pcap_mac_addr_list, wire_mac_addr_list, real_time):
    
        try:
            signal.signal(signal.SIGINT, libs.bt3in.allow_keyboard_interrupt)
            libs.bt3out.print_info("Pcapteller started at %s. Press [CTRL+C] to stop.\n" % (datetime.datetime.now().strftime("%H:%M:%S")))
            libs.bt3out.print_info("Reading \"%s\"..." % (pcap_file))
            pf = self.read_traffic(pcap_file)
            total_packets = len(pf)
            libs.bt3out.print_ok("%s packet(s) found.\n" % (total_packets))
            
            # Traffic manipulation
            packets = self.traffic_manipulation(pf, total_packets, pcap_ip_addr_list, wire_ip_addr_list, pcap_mac_addr_list, wire_mac_addr_list)
            print("")

            # Traffic replay
            if fragmentation:
                libs.bt3out.print_warning("Packet fragmentation enabled. MTU size will be set to %s bytes..." % str(mtu_bytes))
                
            if real_time:
                libs.bt3out.print_warning("Real time enabled. Replay speed will be determined by the actual pcap file...")
            
            libs.bt3out.print_info("Replaying packet(s) via %s..." % (net_interface))
            error = self.traffic_replay(packets, net_interface, real_time, fragmentation, mtu_bytes)
            
            if error:
                libs.bt3out.print_error("Replay incomplete. Packets were injected, but an error ocurred during the replay.")
                libs.bt3out.print_warning("Tips: Try packet fragmentation with different MTU values (e.g. 1024 bytes).\n")

            else:    
                libs.bt3out.print_ok("Replay complete.\n")
            
            libs.bt3out.print_info("Pcapteller finished at %s.\n" % (datetime.datetime.now().strftime("%H:%M:%S")))
            signal.signal(signal.SIGINT, libs.bt3in.prevent_keyboard_interrupt)
        
        except KeyboardInterrupt:
            signal.signal(signal.SIGINT, libs.bt3in.prevent_keyboard_interrupt)
            return True

        except Exception as e:
            libs.bt3out.print_error("%s.\n" % (e))
            return True

